#include <errno.h>

#include <sys/types.h>
#include <sys/fcntl.h>
#include <sys/file.h>
#include <sys/wait.h>

#include <signal.h>
#include <unistd.h>
#include <time.h>
#include <openssl/sha.h>

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#define MAX_USE_TIME    1000
#define MAX_WAIT_TIME   2000
#define MAX_RELAX_TIME  3000

#define CHANNEL         "monochannel"


int count = 3;

int read_random() {
  int fd, r = 0, i;
  char dat[4];

  fd = open("/dev/urandom", O_RDONLY);

  if (fd != -1) {
    read(fd, dat, 4);
    r |= dat[0] & 0xff;
    for(i=1; i<4; i++) {
      r = r << 8;
      r |= dat[i] & 0xff;
    }
  } else {
    perror("open('/dev/urandom')");
    exit(errno);
  }
  close(fd);
  return abs(r);
}

char child_id = 0;

void lock(int fd) {
  struct flock l = { F_WRLCK, SEEK_SET, 0, 0, 0 };
  while (fcntl(fd, F_SETLK, &l) == -1) {
    int r = read_random() % MAX_WAIT_TIME;
    printf("%d    [%d] Channel busy. Wait for %d milliseconds...\n", child_id, getpid(), r);
    (void)usleep(r * 1000);
  }
  printf("%d    [%d] Locked\n", child_id, getpid());
}

void unlock(int fd) {
  struct flock l = { F_UNLCK, SEEK_SET, 0, 0, 0 };  
  if (fcntl(fd, F_SETLK, &l) == -1) {
    printf("%d    [%d] Couldn't unlock (errno=%d)", child_id, getpid(), errno);
  } else {
    printf("%d    [%d] Unlocked\n", child_id, getpid());
  }
}

int get_next_child() {
  return read_random() % (count + 1);
}

void print_sha1(const unsigned char* digest) {
  int i;
  for(i =0; i<SHA_DIGEST_LENGTH; i++) {
    printf("%x", digest[i]); 
  }
}

int check_sha1_digest(const unsigned char* a, const unsigned char* b) {
  int i;
  for(i=0; i<SHA_DIGEST_LENGTH; i++) {
    if (a[i] != b[i]) return -1;
  }
  return 0;
}

void check_inbox(int fd) {
  char cid;
  
  (void) read(fd, &cid, sizeof(char));
  
  if (cid == child_id || cid == count) {
    int message_len = 23;
    unsigned char* message = malloc(message_len);
    const unsigned char* sha_digest;

    if (cid != count) {
      lseek(fd, 2, SEEK_SET);
      (void) read(fd, message, message_len);
      sha_digest = SHA1(message, message_len, NULL);
      printf("Read message '%s'\t", message);
      
      unsigned char received_sha_digest[SHA_DIGEST_LENGTH];
      (void) read(fd, received_sha_digest, SHA_DIGEST_LENGTH);
      if (check_sha1_digest(received_sha_digest, sha_digest) == 0) {
        printf("checksum [");print_sha1(sha_digest);printf("] OK\n");
      } else {
        printf("checksum [");print_sha1(sha_digest);
        printf(", ");
        print_sha1(received_sha_digest);printf("] FAIL\n");
      }
    } else {
      printf("Channel was free\n");
    }

    int next_child = child_id;
    while (next_child == child_id) next_child = get_next_child();
    lseek(fd, 0, SEEK_SET);

    sprintf(message, "pid %6d, child_id %d", getpid(), child_id);
    sha_digest = SHA1(message, message_len, NULL);
        
    write(fd, &next_child, sizeof(char));
    write(fd, &child_id, sizeof(char));
    write(fd, message, message_len);
    write(fd, sha_digest, SHA_DIGEST_LENGTH);

    printf("Write message '%s'\t", message);print_sha1(sha_digest);printf("\n");
    
    free(message);

    int r = read_random() % MAX_USE_TIME;
    printf("%d->%d [%d] Use channel for %d milliseconds...\n", child_id, next_child, getpid(), r);
    usleep(r * 1000);
  } else {
    printf("%d-!%d [%d] Message sent not me...\n", child_id, cid, getpid());
  }
}

int main(int argc, char** argv) {
  if (argc == 2) { 
    count = atoi(argv[1]);
    if (count <= 3) count = 4;
  }

  int channel = open(CHANNEL, O_RDWR|O_CREAT, 0644);
  if (channel == -1) {
    perror("open");
    exit(errno);
  }
  write(channel, &count, sizeof(int));
  close(channel);

  int i;

  pid_t* children = (pid_t*)calloc(count, sizeof(pid_t));
  
  pid_t cpid = 0;
  for(i=0; i<count; i++) {
    child_id = i;
    cpid = fork();
    if (cpid == -1) {
      perror("fork");
      exit(errno);
    } else {
      if (cpid == 0) {
        printf("[%d] Start child %d with pid %d\n", getppid(), child_id, getpid());
        break;
      } else {
        children[i] = cpid;
      }
    }
  }
  if (cpid != 0) {
    int status;
    int pid = wait(&status);
    for (i=0; i<count; i++) {
      kill(children[i], SIGKILL);
    }
    free(children);
  } else { 
    int r;
    for(;;) {
      int fd = open(CHANNEL, O_RDWR);
      if (fd == -1) {
        perror("open");
        exit(errno);
      }

      lock(fd);
      check_inbox(fd);
      unlock(fd);
      close(fd);

      r = read_random() % MAX_RELAX_TIME;
      printf("%d    [%d] Relax after using for %d milliseconds...\n", child_id, getpid(), r);
      usleep(r * 1000);
    }
  }
  return (EXIT_SUCCESS);
} 
